"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Executor = void 0;
const scripts_1 = require("./scripts");
const spawn_process_1 = require("./spawn-process");
const fs = require("fs");
const path = require("path");
const logger_1 = require("./logger");
const common_1 = require("./common");
const glob = require("fast-glob");
const prettyTime = require("pretty-time");
const variables_1 = require("./variables");
const url_1 = require("url");
var Order;
(function (Order) {
    Order[Order["concurrent"] = 0] = "concurrent";
    Order[Order["sequential"] = 1] = "sequential";
})(Order || (Order = {}));
class Executor {
    constructor(shell, environment, settings, scripts, globOptions, confirm, limit, dry, testmode) {
        this.shell = shell;
        this.environment = environment;
        this.settings = settings;
        this.scripts = scripts;
        this.globOptions = globOptions;
        this.confirm = confirm;
        this.testmode = testmode;
        this.startTime = process.hrtime();
        this.limit = limit;
        this.dry = dry;
    }
    execute(scriptInfo) {
        return __awaiter(this, void 0, void 0, function* () {
            const tasks = this.expand(scriptInfo);
            const options = {
                stdio: 'inherit',
                env: this.environment,
                shell: this.shell,
                suppress: false,
                testmode: this.testmode,
                limit: this.limit,
                dry: this.dry
            };
            if (logger_1.Logger.level > 0) {
                logger_1.Logger.info('Script id       :', scriptInfo.name);
                logger_1.Logger.info('Circular        :', Executor.isCircular([tasks]));
                logger_1.Logger.info('Script params   :', scriptInfo.parameters);
                logger_1.Logger.info('Script args     :', scriptInfo.arguments);
                logger_1.Logger.debug('Script object   : ' + (0, common_1.stringify)(scriptInfo.script));
                logger_1.Logger.debug('Script expanded : ' + (0, common_1.stringify)(tasks, Executor.removeEmpties));
                logger_1.Logger.info();
                logger_1.Logger.log();
            }
            if (logger_1.Logger.level > 1) {
                const settings = Object.entries(this.environment).filter(([key, value]) => key.startsWith('launch_setting_'));
                logger_1.Logger.log(common_1.Colors.Bold + 'Launcher Settings Values' + common_1.Colors.Normal);
                logger_1.Logger.log(''.padEnd(process.stdout.columns, '-'));
                for (const [key, value] of settings) {
                    logger_1.Logger.log(common_1.Colors.Dim + key + common_1.Colors.Normal + '=' + common_1.Colors.Green + "'" + value + "'" + common_1.Colors.Normal);
                }
                logger_1.Logger.log(''.padEnd(process.stdout.columns, '-'));
                logger_1.Logger.log('Total: ' + settings.length);
                logger_1.Logger.log();
                logger_1.Logger.log();
                logger_1.Logger.log(common_1.Colors.Bold + 'Launcher Settings Arrays' + common_1.Colors.Normal);
                logger_1.Logger.log(''.padEnd(process.stdout.columns, '-'));
                for (const [key, value] of Object.entries(this.settings.arrays)) {
                    logger_1.Logger.log(common_1.Colors.Dim + key + common_1.Colors.Normal + '=' + (0, common_1.stringify)(value));
                }
                logger_1.Logger.log(''.padEnd(process.stdout.columns, '-'));
                logger_1.Logger.log('Total: ' + settings.length);
                logger_1.Logger.log();
                logger_1.Logger.log();
            }
            const processes = [];
            const concurrent = [...tasks.concurrent];
            const sequential = [...tasks.sequential];
            if (Executor.containsConstraint(tasks)) {
                try {
                    if (yield this.evaluateTask(tasks, options)) {
                        concurrent.push(...tasks['concurrent-then']);
                        sequential.push(...tasks['sequential-then']);
                    }
                    else {
                        concurrent.push(...tasks['concurrent-else']);
                        sequential.push(...tasks['sequential-else']);
                    }
                }
                catch (_a) {
                    return 0;
                }
            }
            processes.push(...(yield this.executeTasks(concurrent, options, Order.concurrent, process.hrtime())));
            processes.push(...(yield this.executeTasks(sequential, options, Order.sequential, process.hrtime())));
            return Executor.wait(processes);
        });
    }
    static isEmptyTask(tasks) {
        if (tasks.confirm.length > 0)
            return false;
        if (tasks.condition.length > 0)
            return false;
        if (tasks.exclusion.length > 0)
            return false;
        if (tasks.concurrent.length > 0)
            return false;
        if (tasks.sequential.length > 0)
            return false;
        if (tasks['concurrent-then'].length > 0)
            return false;
        if (tasks['sequential-then'].length > 0)
            return false;
        if (tasks['concurrent-else'].length > 0)
            return false;
        if (tasks['sequential-else'].length > 0)
            return false;
        return true;
    }
    static convertSingleQuote(command) {
        const argv = (0, common_1.stringToArgv)(command);
        const result = [];
        for (let value of argv) {
            if (value.includes(' '))
                value = '"' + value + '"';
            result.push(value);
        }
        return result.join(' ');
    }
    static removeEmpties(key, value) {
        if (value instanceof Array && value.length === 0)
            return undefined;
        return value;
    }
    static containsConstraint(task) {
        return task.condition.length > 0 || task.exclusion.length > 0 || task.confirm.length > 0;
    }
    static expandArguments(text, args) {
        const variables = [];
        for (let index = 0; index < args.length; index++) {
            variables.push([index + '\\*', args.slice(index).join(' ')]);
            variables.push([index.toString(), args[index]]);
        }
        variables.push(['\\*', args.slice(1).join(' ')]);
        text = variables_1.Variables.expand(text, variables);
        return variables_1.Variables.remove(text, '\\d+');
    }
    static expandEnvironment(text, environment) {
        return variables_1.Variables.expand(text, Object.entries(environment));
    }
    static removeEnvironment(text) {
        return variables_1.Variables.remove(text, '\\w+');
    }
    static extendEnvironment(environment, command) {
        const match = command.trim().match(Executor.assignmentPattern);
        if (match !== null) {
            if (match[2].startsWith('$')) {
                const environmentValue = match[2].replace('$', '');
                const settings = Object.entries(environment).filter(([key]) => key.startsWith(environmentValue + '_'));
                for (const [key, value] of settings) {
                    if (key.startsWith(environmentValue)) {
                        const environmentKey = match[1] + key.replace(environmentValue, '');
                        environment[environmentKey] = value;
                        logger_1.Logger.debug(common_1.Colors.Bold + 'Set environment' + common_1.Colors.Normal + ' : ' + common_1.Colors.Green + "'" + environmentKey + '=' + value + "'" + common_1.Colors.Normal);
                    }
                }
            }
            else {
                environment[match[1]] = match[2];
                logger_1.Logger.debug(common_1.Colors.Bold + 'Set environment' + common_1.Colors.Normal + ' : ' + common_1.Colors.Green + "'" + match[1] + '=' + match[2] + "'" + common_1.Colors.Normal);
            }
            return true;
        }
        return false;
    }
    static isDynamicPattern(item, options) {
        if (!item || item.match(/^['|"].*['|"]$/) !== null)
            return false;
        return glob.isDynamicPattern(item.replace(/\\./g, ''), options);
    }
    static splitString(value) {
        const result = [];
        let last = 0;
        let index = 0;
        while (index < value.length) {
            const char = value[index];
            if (char === ' ') {
                result.push(value.substr(last, index - last));
                last = index + 1;
            }
            if (char === '"') {
                while (++index < value.length && value[index] !== '"')
                    ;
            }
            if (char === "'") {
                while (++index < value.length && value[index] !== "'")
                    ;
            }
            index++;
        }
        result.push(value.substr(last, index - last));
        return result;
    }
    static expandGlobs(pattern, options) {
        const result = [];
        for (const item of Executor.splitString(pattern)) {
            let value = [item];
            if (Executor.isDynamicPattern(item, options)) {
                value = glob.sync(item, options);
                if (options.nonull && value.length === 0)
                    value = [item];
            }
            result.push(...value);
        }
        return result.join(' ');
    }
    static getCommandInfo(command, options) {
        if (!command)
            command = '';
        options = Object.assign({}, options);
        options.env = Object.assign({}, options.env);
        command = Executor.expandEnvironment(command, options.env);
        if (!options.cwd)
            options.cwd = '';
        let args = [];
        if (!options.shell) {
            args = (0, common_1.stringToArgv)(command);
            command = args[0];
            args.shift();
        }
        if (command === '') {
            console.log();
            return { command: null, args, options };
        }
        if (command === 'echo') {
            if (process.platform === 'win32')
                command += '.';
            return { command, args, options };
        }
        if (command === '--') {
            if (options.testmode) {
                console.log(''.padEnd(32, '-'));
            }
            else {
                console.log(''.padEnd(process.stdout.columns, '-'));
            }
            return { command: null, args, options };
        }
        if (command.startsWith('#')) {
            logger_1.Logger.log(common_1.Colors.Bold + 'Skipping action' + common_1.Colors.Normal + ' : ' + common_1.Colors.Green + "'" + command + "'" + common_1.Colors.Normal, args);
            return { command: null, args, options };
        }
        if (command.endsWith(' || true')) {
            command = command.replace(/ \|\| true$/, '');
            options.suppress = true;
        }
        if (Executor.extendEnvironment(options.env, command))
            return { command: null, args, options };
        if (options.cwd instanceof url_1.URL) {
            throw new Error('A directory type of type URL is not supported.');
        }
        const fullPath = path.join(path.resolve(options.cwd), command);
        if (fs.existsSync(fullPath)) {
            options.cwd = fullPath;
            return { command: null, args, options };
        }
        return { command, args, options };
    }
    static wait(processes) {
        return __awaiter(this, void 0, void 0, function* () {
            let exitCode = 0;
            for (const item of processes) {
                if (item instanceof Promise) {
                    exitCode += yield this.wait(yield item);
                }
                else {
                    exitCode += yield item.wait();
                }
            }
            return exitCode;
        });
    }
    static isCircular(tasks) {
        for (const task of tasks) {
            if (typeof task !== 'string') {
                if (task.circular)
                    return true;
                if (Executor.isCircular(task.concurrent))
                    return true;
                if (Executor.isCircular(task.sequential))
                    return true;
                if (Executor.isCircular(task['concurrent-then']))
                    return true;
                if (Executor.isCircular(task['sequential-then']))
                    return true;
                if (Executor.isCircular(task['concurrent-else']))
                    return true;
                if (Executor.isCircular(task['sequential-else']))
                    return true;
            }
        }
        return false;
    }
    preprocessScripts(scripts) {
        const result = [];
        if (scripts) {
            if (typeof scripts !== 'string') {
                result.push(...scripts);
            }
            else {
                result.push(scripts);
            }
        }
        return result;
    }
    expand(scriptInfo, parents = []) {
        const script = scriptInfo.script;
        const repeater = script.repeater;
        const result = {
            'circular': false,
            'parameters': {},
            'confirm': [],
            'condition': [],
            'exclusion': [],
            'concurrent': [],
            'sequential': [],
            'concurrent-then': [],
            'sequential-then': [],
            'concurrent-else': [],
            'sequential-else': []
        };
        const environment = Object.assign(Object.assign({}, scriptInfo.parameters), this.settings.values);
        for (const [name, value] of Object.entries(scriptInfo.parameters)) {
            if (value === undefined) {
                throw new Error('The parameter "' + name + '" of script "' + scriptInfo.name + '" is ' + value + '.');
            }
        }
        if (!repeater) {
            const confirm = [];
            const concurrent = [];
            const sequential = [];
            const concurrentThen = [];
            const sequentialThen = [];
            const concurrentElse = [];
            const sequentialElse = [];
            if (script instanceof Array)
                (scriptInfo.inline ? concurrent : sequential).push(...script);
            if (typeof script === 'string')
                sequential.push(script);
            parents = [...parents, scriptInfo.name];
            confirm.push(...this.preprocessScripts(script.confirm));
            concurrent.push(...this.preprocessScripts(script.concurrent));
            sequential.push(...this.preprocessScripts(script.sequential));
            concurrentThen.push(...this.preprocessScripts(script['concurrent-then']));
            sequentialThen.push(...this.preprocessScripts(script['sequential-then']));
            concurrentElse.push(...this.preprocessScripts(script['concurrent-else']));
            sequentialElse.push(...this.preprocessScripts(script['sequential-else']));
            result.parameters = scriptInfo.parameters;
            result.confirm.push(...this.expandConstraint(parents, script.confirm, environment, scriptInfo.arguments, result));
            result.condition.push(...this.expandConstraint(parents, script.condition, environment, scriptInfo.arguments, result));
            result.exclusion.push(...this.expandConstraint(parents, script.exclusion, environment, scriptInfo.arguments, result));
            result.concurrent.push(...this.expandTasks(parents, concurrent, environment, scriptInfo.arguments, scriptInfo.parameters, result));
            result.sequential.push(...this.expandTasks(parents, sequential, environment, scriptInfo.arguments, scriptInfo.parameters, result));
            result['concurrent-then'].push(...this.expandTasks(parents, concurrentThen, environment, scriptInfo.arguments, scriptInfo.parameters, result));
            result['sequential-then'].push(...this.expandTasks(parents, sequentialThen, environment, scriptInfo.arguments, scriptInfo.parameters, result));
            result['concurrent-else'].push(...this.expandTasks(parents, concurrentElse, environment, scriptInfo.arguments, scriptInfo.parameters, result));
            result['sequential-else'].push(...this.expandTasks(parents, sequentialElse, environment, scriptInfo.arguments, scriptInfo.parameters, result));
        }
        else {
            let array = repeater.replace(/^\$/, '');
            array = Executor.expandArguments(array, scriptInfo.arguments);
            array = Executor.expandEnvironment(array, environment);
            const settings = this.settings.arrays[array] || [];
            const repeaterTask = {
                name: scriptInfo.name,
                inline: scriptInfo.inline,
                multiple: false,
                parameters: null,
                arguments: scriptInfo.arguments,
                script: Object.assign({}, scriptInfo.script)
            };
            repeaterTask.script.repeater = null;
            for (const setting of settings) {
                repeaterTask.parameters = Object.assign(Object.assign({}, scriptInfo.parameters), setting);
                const task = this.expand(repeaterTask);
                result.sequential.push({
                    'circular': false,
                    'parameters': repeaterTask.parameters,
                    'confirm': task.confirm,
                    'condition': task.condition,
                    'exclusion': task.exclusion,
                    'concurrent': task.concurrent,
                    'sequential': task.sequential,
                    'concurrent-then': task['concurrent-then'],
                    'sequential-then': task['sequential-then'],
                    'concurrent-else': task['concurrent-else'],
                    'sequential-else': task['sequential-else']
                });
            }
        }
        return result;
    }
    expandConstraint(parents, constraints, environment, args, meta) {
        const result = [];
        if (!constraints)
            return result;
        if (typeof constraints === 'string')
            constraints = [constraints];
        if (!(constraints instanceof Array))
            throw new Error('Constraint object value not supported: ' + (0, common_1.stringify)(constraints));
        for (let constraint of constraints) {
            constraint = Executor.expandArguments(constraint, args);
            constraint = Executor.expandEnvironment(constraint, environment);
            const scripts = this.scripts.find(constraint);
            const scriptInfo = scripts_1.Scripts.select(scripts, parents, meta);
            if (scriptInfo) {
                if (scriptInfo.multiple) {
                    for (let script of scriptInfo.script) {
                        script = Executor.expandArguments(script, args);
                        script = Executor.expandEnvironment(script, environment);
                        result.push(script);
                    }
                    continue;
                }
                scriptInfo.arguments = [scriptInfo.name, ...scriptInfo.arguments];
                result.push(...this.expandConstraint([...parents, constraint], scriptInfo.script, Object.assign(Object.assign({}, this.environment), scriptInfo.parameters), args, meta));
            }
            else {
                result.push(constraint);
            }
        }
        return result;
    }
    executeTasks(tasks, options, order, startTime, limiter = new common_1.Limiter(options.limit)) {
        return __awaiter(this, void 0, void 0, function* () {
            const processes = [];
            const suppress = options.suppress;
            for (const task of tasks) {
                options.suppress = suppress;
                if (typeof task === 'string') {
                    options.env.launch_time_current = (0, common_1.formatTime)();
                    options.env.launch_time_elapsed = prettyTime(process.hrtime(startTime), 'ms');
                    if (options.testmode) {
                        options.env.launch_time_current = (0, common_1.formatTime)(new Date('2019-09-16T12:33:42.285').getTime(), 0);
                        options.env.launch_time_elapsed = '137ms';
                    }
                    const info = Executor.getCommandInfo(task, options);
                    options = info.options;
                    if (info.command) {
                        if (options.cwd instanceof url_1.URL) {
                            throw new Error('A directory type of type URL is not supported.');
                        }
                        let command = Executor.expandGlobs(info.command, Object.assign(Object.assign({}, this.globOptions), { cwd: options.cwd }));
                        command = Executor.removeEnvironment(command);
                        // Remove environment,argument and glob escapings
                        command = command.replace(/\\(.)/g, '$1');
                        if (process.platform === 'win32') {
                            command = Executor.convertSingleQuote(command);
                            if (command.startsWith('echo'))
                                command = 'echo' + command.replace('echo', '').replace(/^\s*\"(.*)\"\s*$/g, ' $1');
                        }
                        logger_1.Logger.log(common_1.Colors.Bold + 'Spawn action   ' + common_1.Colors.Normal + ' : ' + common_1.Colors.Green + "'" + command + "'" + common_1.Colors.Normal, info.args);
                        logger_1.Logger.log('Spawn options   : { order=' + common_1.Colors.Cyan + Order[order] + common_1.Colors.Normal + ', supress=' + common_1.Colors.Yellow + options.suppress + common_1.Colors.Normal + ' }');
                        yield limiter.enter();
                        const commandProcess = spawn_process_1.Process.spawn(command, info.args, options);
                        processes.push(commandProcess);
                        if (order === Order.sequential) {
                            const exitCode = yield commandProcess.wait();
                            limiter.leave();
                            if (exitCode !== 0)
                                break;
                        }
                        else {
                            commandProcess.wait().then(() => limiter.leave());
                        }
                    }
                }
                else {
                    const concurrent = [...task.concurrent];
                    const sequential = [...task.sequential];
                    if (Executor.containsConstraint(task)) {
                        try {
                            if (yield this.evaluateTask(task, options)) {
                                concurrent.push(...task['concurrent-then']);
                                sequential.push(...task['sequential-then']);
                            }
                            else {
                                concurrent.push(...task['concurrent-else']);
                                sequential.push(...task['sequential-else']);
                            }
                        }
                        catch (_a) {
                            return [];
                        }
                    }
                    const concurrentProcesses = this.executeTasks(concurrent, options, Order.concurrent, process.hrtime(), limiter);
                    const sequentialProcesses = this.executeTasks(sequential, options, Order.sequential, process.hrtime(), limiter);
                    processes.push(concurrentProcesses);
                    processes.push(sequentialProcesses);
                    if (order === Order.sequential) {
                        let exitCode = 0;
                        exitCode += yield Executor.wait(yield concurrentProcesses);
                        exitCode += yield Executor.wait(yield sequentialProcesses);
                        if (exitCode > 0)
                            break;
                    }
                }
            }
            return processes;
        });
    }
    evaluateConstraint(type, command, options, outputPattern) {
        return __awaiter(this, void 0, void 0, function* () {
            if (outputPattern)
                logger_1.Logger.log('Grep pattern    : ' + common_1.Colors.Green + "'" + outputPattern + "'" + common_1.Colors.Normal);
            options = Object.assign({}, options);
            command = command.replace(/require\s*\(\s*'\.\//g, "require(process.cwd() + '/");
            try {
                const result = eval(command);
                if (typeof result !== 'boolean')
                    throw new Error('type not supported');
                logger_1.Logger.log(common_1.Colors.Bold + type + '       : ' + common_1.Colors.Normal + common_1.Colors.Green + "'" + command + "'" + common_1.Colors.Normal);
                logger_1.Logger.log('Result          : ' + result);
                logger_1.Logger.log();
                logger_1.Logger.log();
                return result;
            }
            catch (error) {
                // Not a valid javascript expression, continue
            }
            if (process.platform === 'win32') {
                command = Executor.convertSingleQuote(command);
                if (command.startsWith('echo'))
                    command = 'echo' + command.replace('echo', '').replace(/^\s*\"(.*)\"\s*$/g, ' $1');
            }
            logger_1.Logger.log(common_1.Colors.Bold + type + '       : ' + common_1.Colors.Normal + common_1.Colors.Green + "'" + command + "'" + common_1.Colors.Normal);
            if (options.cwd instanceof url_1.URL) {
                throw new Error('A directory type of type URL is not supported.');
            }
            if (fs.existsSync(path.join(path.resolve(options.cwd), command))) {
                logger_1.Logger.log(''.padEnd(process.stdout.columns, '-'));
                logger_1.Logger.log(common_1.Colors.Dim + 'directory exists' + common_1.Colors.Normal);
                logger_1.Logger.log(''.padEnd(process.stdout.columns, '-'));
                logger_1.Logger.log('Result          : true');
                logger_1.Logger.log();
                logger_1.Logger.log();
                return true;
            }
            try {
                options.stdio = ['inherit', 'pipe', 'pipe'];
                if (outputPattern) {
                    options.extraLogInfo = process => {
                        const matches = (process.stdout + process.stderr).match(outputPattern);
                        return 'grep=' + (matches === null ? 'failed' : 'success');
                    };
                }
                const commandProcess = spawn_process_1.Process.spawn(command, [], options);
                const exitCode = yield commandProcess.wait();
                if (outputPattern)
                    return (commandProcess.stdout + commandProcess.stderr).match(outputPattern) != null;
                return exitCode === 0;
            }
            catch (_a) {
                return false;
            }
        });
    }
    evaluateTask(task, options) {
        return __awaiter(this, void 0, void 0, function* () {
            let condition = true;
            let exclusion = false;
            let confirmation = !(task.confirm.length > 0);
            options = Object.assign({}, options);
            if (!options.cwd)
                options.cwd = '';
            for (let constraint of task.condition) {
                constraint = Executor.expandEnvironment(constraint, options.env);
                if (!Executor.extendEnvironment(options.env, constraint)) {
                    const outputMatches = constraint.match(/(.*)\|\?(.*)/);
                    let outputPattern = null;
                    if (outputMatches !== null) {
                        constraint = outputMatches[1].trim();
                        outputPattern = Executor.expandEnvironment(outputMatches[2].trim(), options.env);
                        outputPattern = Executor.removeEnvironment(outputPattern);
                    }
                    if (options.cwd instanceof url_1.URL) {
                        throw new Error('A directory type of type URL is not supported.');
                    }
                    constraint = Executor.expandGlobs(constraint, Object.assign(Object.assign({}, this.globOptions), { cwd: options.cwd }));
                    constraint = Executor.removeEnvironment(constraint);
                    // Remove environment and argument escaping
                    constraint = constraint.replace(/\\\$/g, '$');
                    if (!(yield this.evaluateConstraint('Condition', constraint, options, outputPattern))) {
                        condition = false;
                        break;
                    }
                }
            }
            for (let constraint of task.exclusion) {
                constraint = Executor.expandEnvironment(constraint, options.env);
                if (!Executor.extendEnvironment(options.env, constraint)) {
                    const outputMatches = constraint.match(/(.*)\|\?(.*)/);
                    let outputPattern = null;
                    if (outputMatches !== null) {
                        constraint = outputMatches[1].trim();
                        outputPattern = Executor.expandEnvironment(outputMatches[2].trim(), options.env);
                        outputPattern = Executor.removeEnvironment(outputPattern);
                    }
                    if (options.cwd instanceof url_1.URL) {
                        throw new Error('A directory type of type URL is not supported.');
                    }
                    constraint = Executor.expandGlobs(constraint, Object.assign(Object.assign({}, this.globOptions), { cwd: options.cwd }));
                    constraint = Executor.removeEnvironment(constraint);
                    if (yield this.evaluateConstraint('Exclusion', constraint, options, outputPattern)) {
                        exclusion = true;
                        break;
                    }
                }
            }
            for (let confirm of task.confirm) {
                confirm = Executor.expandEnvironment(confirm, options.env);
                if (options.cwd instanceof url_1.URL) {
                    throw new Error('A directory type of type URL is not supported.');
                }
                confirm = Executor.expandGlobs(confirm, Object.assign(Object.assign({}, this.globOptions), { cwd: options.cwd }));
                confirm = Executor.removeEnvironment(confirm);
                logger_1.Logger.log(common_1.Colors.Bold + 'Confirm         : ' + common_1.Colors.Normal + common_1.Colors.Green + "'" + confirm + "'" + common_1.Colors.Normal);
                confirmation = yield (0, common_1.confirmPrompt)(confirm, this.confirm);
                if (!confirmation)
                    break;
            }
            return condition && confirmation && !exclusion;
        });
    }
    expandTasks(parents, tasks, environment, args, parameters, meta) {
        const result = [];
        for (let task of tasks) {
            if (typeof task === 'string') {
                task = Executor.expandArguments(task, args);
                task = Executor.expandEnvironment(task, environment);
                const scripts = this.scripts.find(task);
                const scriptInfo = scripts_1.Scripts.select(scripts, parents, meta);
                if (scriptInfo) {
                    if (scriptInfo.multiple) {
                        for (let script of scriptInfo.script) {
                            script = Executor.expandArguments(script, args);
                            script = Executor.expandEnvironment(script, environment);
                            result.push(script);
                        }
                        continue;
                    }
                    scriptInfo.arguments = [scriptInfo.name, ...scriptInfo.arguments];
                    const tasks = this.expand(scriptInfo, parents);
                    if (!Executor.isEmptyTask(tasks))
                        result.push(tasks);
                }
                else {
                    if (tasks.length === 1 && parents.length > 0 && task === parents[parents.length - 1])
                        task = Executor.expandArguments(task + ' $*', args);
                    result.push(task);
                }
            }
            else {
                const scriptInfo = {
                    name: 'inline-script-block',
                    inline: true,
                    multiple: false,
                    parameters: parameters,
                    arguments: args,
                    script: task
                };
                if (task.repeater) {
                    result.push(...this.expand(scriptInfo, parents).sequential);
                }
                else {
                    result.push(this.expand(scriptInfo, parents));
                }
            }
        }
        return result;
    }
}
exports.Executor = Executor;
Executor.assignmentPattern = `^(\\w+\)=([\\$\\w\\,\\.\\-\\@\\#\\%\\^\\*\\:\\;\\+\\/\\\\~\\=\\[\\]\\{\\}\\"\\']+|\".*\"|\'.*\')$`;
