"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Scripts = void 0;
const common_1 = require("./common");
class Scripts {
    constructor(scripts) {
        this.scripts = scripts;
    }
    static select(scripts, filters = null, meta = null) {
        const length = scripts.length;
        if (!filters)
            filters = [];
        if (!meta)
            meta = { circular: false };
        if (length > 0) {
            if (scripts.some(item => item.multiple)) {
                scripts = scripts.filter(Scripts.containsScript.bind(filters));
                if (scripts.length !== length)
                    meta.circular = true;
                const resolvedScripts = Scripts.resolveScripts(scripts);
                return {
                    name: null,
                    inline: false,
                    multiple: true,
                    parameters: {},
                    arguments: [],
                    script: resolvedScripts
                };
            }
            // Sort by number of values and parameters
            scripts = scripts.sort((itemA, itemB) => (Scripts.countValues(itemB.name) - Scripts.countValues(itemA.name)) * 50 + (Scripts.countParams(itemA.name) - Scripts.countParams(itemB.name)));
            for (const script of scripts) {
                if (!filters.includes(script.name)) {
                    return script;
                }
                else {
                    meta.circular = true;
                }
            }
        }
        return null;
    }
    static parse(pattern) {
        const args = (0, common_1.stringToArgv)(pattern);
        return {
            command: args.length > 0 ? args[0] : '',
            arguments: args.slice(1)
        };
    }
    find(...patterns) {
        const scripts = [];
        const multiple = patterns.length > 1;
        for (const pattern of patterns) {
            const info = Scripts.parse(pattern);
            for (const [name, script] of Object.entries(this.scripts)) {
                const parameters = Scripts.getParameters(name, info.command, multiple);
                if (parameters !== null) {
                    scripts.push({
                        name: name,
                        inline: false,
                        multiple: multiple,
                        parameters: parameters,
                        arguments: info.arguments,
                        script: script
                    });
                }
            }
            if (scripts.length === 0) {
                for (const [name, script] of Object.entries(this.scripts)) {
                    const parameters = Scripts.getParameters(name, info.command, true);
                    if (parameters !== null) {
                        scripts.push({
                            name: name,
                            inline: false,
                            multiple: true,
                            parameters: parameters,
                            arguments: info.arguments,
                            script: script
                        });
                    }
                }
            }
        }
        return scripts;
    }
    static containsScript(script) {
        const filters = this || [];
        return !filters.some(filter => Scripts.getParameters(script.name, filter, true) !== null);
    }
    static resolveScripts(scripts) {
        const result = [];
        for (const item of scripts) {
            if (item.script instanceof Array) {
                result.push(...item.script);
            }
            else {
                result.push(item.script);
            }
        }
        return result;
    }
    static countParams(name) {
        const columns = name.split(':');
        return columns.filter(item => item.startsWith('$')).length;
    }
    static countValues(name) {
        const columns = name.split(':');
        return columns.length - columns.filter(item => item.startsWith('$')).length;
    }
    static getParameters(signature, reference, wildcard) {
        const signatureParams = signature.split(':');
        const referenceParams = reference.split(':');
        const parameters = {};
        let defaultParameters = 0;
        if (signatureParams.length < referenceParams.length)
            return null;
        for (let index = 0; index < signatureParams.length; index++) {
            const signatureParam = signatureParams[index];
            const referenceParam = referenceParams[index];
            if (signatureParam.trim().startsWith('$')) {
                const columns = signatureParam.trim().substr(1).split('=');
                const name = columns[0];
                parameters[name] = referenceParam;
                if (!referenceParam) {
                    parameters[name] = columns[1];
                    if (columns[1] !== undefined && referenceParam === undefined)
                        defaultParameters++;
                }
                continue;
            }
            if (wildcard && referenceParam === '*')
                continue;
            if (signatureParam !== referenceParam)
                return null;
        }
        if (signatureParams.length !== referenceParams.length + defaultParameters)
            return null;
        return parameters;
    }
}
exports.Scripts = Scripts;
